/**
  *  \file planetscentral/dbclient/dbhash.cc
  */

#include "planetscentral/dbclient/dbhash.h"
#include "planetscentral/dbclient/dbclient.h"
#include "int/array.h"

DbHash::DbHash(DbClient& client, string_t name)
    : DbKey(client, name)
{ }

bool
DbHash::removeField(string_t name)
{
    return client.callInt(DbRequest().withString("HDEL").withString(key).withString(name));
}

bool
DbHash::hasField(string_t name)
{
    return client.callInt(DbRequest().withString("HEXISTS").withString(key).withString(name));
}

string_t
DbHash::getStringField(string_t name)
{
    return client.callString(DbRequest().withString("HGET").withString(key).withString(name));
}

int32_t
DbHash::getIntField(string_t name)
{
    return client.callInt(DbRequest().withString("HGET").withString(key).withString(name));
}

IntValue*
DbHash::getFieldRaw(string_t name)
{
    return client.call(DbRequest().withString("HGET").withString(key).withString(name));
}

int32_t
DbHash::addToField(string_t name, int32_t n)
{
    return client.callInt(DbRequest().withString("HINCRBY").withString(key).withString(name).withInt(n));
}

int32_t
DbHash::size()
{
    return client.callInt(DbRequest().withString("HLEN").withString(key));
}

bool
DbHash::setField(string_t name, string_t value)
{
    return client.callInt(DbRequest().withString("HSET").withString(key).withString(name).withString(value));
}

bool
DbHash::setField(string_t name, int32_t value)
{
    return client.callInt(DbRequest().withString("HSET").withString(key).withString(name).withInt(value));
}

bool
DbHash::setFieldRaw(string_t name, IntValue* value)
{
    return client.callInt(DbRequest().withString("HSET").withString(key).withString(name).withRawValue(value));
}

void
DbHash::setFields(const StringList_t& vec)
{
    client.callVoid(DbRequest().withString("HMSET").withString(key).withInlineStringList(vec));
}

void
DbHash::getAll(StringList_t& vec)
{
    std::auto_ptr<IntValue> val(client.call(DbRequest().withString("HGETALL").withString(key)));
    toStringListResult(val.get(), vec);
}

void
DbHash::getAll(IntList_t& vec)
{
    std::auto_ptr<IntValue> val(client.call(DbRequest().withString("HGETALL").withString(key)));
    toIntListResult(val.get(), vec);
}

void
DbHash::setAll(const StringList_t& vec)
{
    return client.callVoid(DbRequest().withString("HMSET").withString(key).withInlineStringList(vec));
}

void
DbHash::keys(StringList_t& vec)
{
    std::auto_ptr<IntValue> val(client.call(DbRequest().withString("HKEYS").withString(key)));
    toStringListResult(val.get(), vec);
}

IntValue*
DbHash::getFields(const StringList_t& vec)
{
    if (vec.empty()) {
        Ptr<IntArrayData> ad = new IntArrayData();
        ad->addDimension(0);
        return new IntArray(ad);
    } else {
        return client.call(DbRequest().withString("HMGET").withString(key).withInlineStringList(vec));
    }
}

DbHashIntField
DbHash::intField(string_t name)
{
    return DbHashIntField(*this, name);
}

DbHashStringField
DbHash::stringField(string_t name)
{
    return DbHashStringField(*this, name);
}

/****************************** DbHashField ******************************/

DbHashField::DbHashField(DbHash& h, const string_t& name)
    : h(h), name(name)
{ }

DbHashField::~DbHashField()
{ }

bool
DbHashField::exists()
{
    return h.hasField(name);
}

bool
DbHashField::remove()
{
    return h.removeField(name);
}

IntValue*
DbHashField::getRaw()
{
    return h.getFieldRaw(name);
}

/***************************** DbHashIntField ****************************/

DbHashIntField::DbHashIntField(DbHash& h, const string_t& name)
    : DbHashField(h, name)
{ }

int32_t
DbHashIntField::operator++()
{
    return (*this) += 1;
}

int32_t
DbHashIntField::operator+=(int32_t incr)
{
    return h.addToField(name, incr);
}

int32_t
DbHashIntField::operator--()
{
    return (*this) += -1;
}

int32_t
DbHashIntField::operator-=(int32_t incr)
{
    return (*this) += -incr;
}

bool
DbHashIntField::set(int32_t value)
{
    return h.setField(name, value);
}

int32_t
DbHashIntField::get()
{
    return h.getIntField(name);
}

/*************************** DbHashStringField ***************************/

DbHashStringField::DbHashStringField(DbHash& h, const string_t& name)
    : DbHashField(h, name)
{ }

bool
DbHashStringField::set(string_t value)
{
    return h.setField(name, value);
}

string_t
DbHashStringField::get()
{
    return h.getStringField(name);
}
