/**
  *  \file planetscentral/host/cmdplayer.cc
  *  \brief Host: PLAYER commands
  */

#include <cstdio>
#include "planetscentral/host/cmdplayer.h"
#include "cpluslib/format.h"
#include "game/player.h"
#include "game/struct.h"
#include "game/trn.h"
#include "int/args.h"
#include "int/array.h"
#include "io/memstream.h"
#include "planetscentral/dbclient/fileclient.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/host/cron.h"
#include "planetscentral/host/db.h"
#include "planetscentral/host/exec.h"
#include "planetscentral/host/host.h"
#include "planetscentral/host/install.h"
#include "planetscentral/host/lock.h"
#include "planetscentral/host/runner.h"
#include "planetscentral/host/schedule.h"
#include "util/misc.h"
#include "planetscentral/host/ranking.h"
#include "io/io.h"

static uint32_t
indexOf(const StringList_t& list, const string_t& ele)
{
    uint32_t i = 0;
    uint32_t size = list.size();
    while (i < size && list[i] != ele) {
        ++i;
    }
    return i;
}


/** PLAYERJOIN command. */
void
doPlayerAdd(HostConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCount(3);
    int32_t gameId = toIntResult(args.getNext());
    int32_t slotId = toIntResult(args.getNext());
    string_t userId = toStringResult(args.getNext());

    // Lock game
    GameLock gl(gameId);
    gl.lockOrDie(IntentDB);

    // Execute
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);

    DbSubtree user(database_connection, USER_ROOT);
    if (!user.stringSetKey("all").contains(userId)) {
        conn.sendError("404 User does not exist");
        return;
    }

    // Check preconditions

    /* Only joining/running games can be joined */
    string_t gameState = game.getState();
    if (gameState != "joining" && gameState != "running") {
        conn.sendError("412 Wrong game state");
        return;
    }

    /* Admin can join anyone, anyone can join unlisted/public as himself,
       owner can put anyone into his games */
    string_t gameType = game.getType();
    bool isAdminAccess = (conn.isAdmin() || conn.getUser() == game.getOwner());
    if (!(isAdminAccess
          || ((gameType == "public" || gameType == "unlisted") && conn.getUser() == userId)))
    {
        conn.sendError("403 Permission denied");
        return;
    }

    /* User must not already be playing elsewhere */
    if (!isAdminAccess && game.isUserOnGameAsPrimary(userId) && !game.isMultiJoinAllowed()) {
        conn.sendError("403 Permission denied");
        return;
    }

    /* Slot must be empty */
    if (!game.isSlotInGame(slotId) || game.isSlotPlayed(slotId)) {
        conn.sendError("409 Slot is not available");
        return;
    }

    /* All conditions fulfilled */
    game.pushPlayerSlot(slotId, userId, user);
    game.addUserHistoryItem(userId == conn.getUser() ? "game-join" : "game-join-other",
                            format("%s:%d", userId, slotId),
                            user, userId);
    if (!game.hasAnyOpenSlot()) {
        game.setConfigInt("lastPlayerJoined", getCurrentTime());
        cron.handleGameStart(gameId);
    }
    conn.sendSuccess("OK");
}

/** PLAYERSUBST command. */
void
doPlayerSubst(HostConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCount(3);
    int32_t gameId = toIntResult(args.getNext());
    int32_t slotId = toIntResult(args.getNext());
    string_t userId = toStringResult(args.getNext());

    // Lock game
    GameLock gl(gameId);
    gl.lockOrDie(IntentDB);

    // Execute
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    DbSubtree user(database_connection, USER_ROOT);

    // Check preconditions

    /* Only joining/running games */
    string_t gameState = game.getState();
    if (gameState != "joining" && gameState != "running") {
        conn.sendError("412 Wrong game state");
        return;
    }

    /* Check player list */
    StringList_t players;
    game.listPlayers(slotId, players);
    uint32_t numPlayers = players.size();
    if (numPlayers == 0) {
        /* If list is empty, they can use PLAYERJOIN instead, which also
           checks that the slot actually exists */
        conn.sendError("412 Slot not in use");
        return;
    }

    uint32_t userIndex = indexOf(players, userId);
    if (conn.isAdmin() || conn.getUser() == game.getOwner()) {
        /* Admin version: if player is on list, drop all its replacements; otherwise, add him,
           no more questions asked. */
        if (userIndex >= numPlayers) {
            /* Not on list */
            game.pushPlayerSlot(slotId, userId, user);
        } else {
            /* Is on list */
            while (numPlayers > userIndex+1) {
                game.popPlayerSlot(slotId, user);
                --numPlayers;
            }
        }
    } else {
        /* User version: caller must be on list. */
        uint32_t callerIndex = indexOf(players, conn.getUser());
        if (callerIndex >= numPlayers || userIndex < callerIndex) {
            /* Caller not on list, or user on list before caller */
            conn.sendError("403 Permission denied");
            return;
        }
        while (numPlayers > callerIndex+1) {
            game.popPlayerSlot(slotId, user);
            --numPlayers;
        }
        game.pushPlayerSlot(slotId, userId, user);
    }
    game.addUserHistoryItem("game-subst", format("%s:%d", userId, slotId), user, userId);

    conn.sendSuccess("OK");
}

/** PLAYERRESIGN command. */
void
doPlayerResign(HostConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCount(3);
    int32_t gameId = toIntResult(args.getNext());
    int32_t slotId = toIntResult(args.getNext());
    string_t userId = toStringResult(args.getNext());

    // Lock game
    GameLock gl(gameId);
    gl.lockOrDie(IntentDB);

    // Execute
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    DbSubtree user(database_connection, USER_ROOT);

    // Check preconditions

    /* Only joining/running games */
    string_t gameState = game.getState();
    if (gameState != "joining" && gameState != "running") {
        conn.sendError("412 Wrong game state");
        return;
    }

    /* User must be in the game */
    StringList_t players;
    game.listPlayers(slotId, players);
    uint32_t numPlayers = players.size();
    uint32_t userIndex = indexOf(players, userId);
    if (userIndex >= numPlayers) {
        conn.sendError("403 Permission denied");
        return;
    }

    /* Caller must be admin, in the list before the user, or owner of the game */
    uint32_t callerIndex = indexOf(players, conn.getUser());
    if (callerIndex > userIndex && !conn.isAdmin() && game.getOwner() != conn.getUser()) {
        conn.sendError("403 Permission denied");
        return;
    }

    /* OK */
    while (numPlayers > userIndex) {
        game.popPlayerSlot(slotId, user);
        --numPlayers;
    }

    /* If we've made a slot empty, notify the scheduler.
       - if this game is joining, this will stop the Master run
       - if this game is running and in run-when-all-turns-are-in mode,
         host may run because all other turns are in. */
    if (userIndex == 0) {
        /* Is this slot dead now? If so, drop it. */
        bool dead = false;
        if (gameState == "running") {
            string_t packedScore = game.getGame().subtree("turn").subtree(game.getConfigInt("turn")).hashKey("scores").getStringField("timscore");
            if (slotId <= 0 || packedScore.size() < slotId*4U || getInt32(packedScore.data() + 4*(slotId-1)) <= 0) {
                dead = true;
            }
        }
        if (dead) {
            game.getSlot(slotId).slotStatus().set(0);
        }

        /* History */
        game.addUserHistoryItem(userId == conn.getUser() ? dead ? "game-resign-dead" : "game-resign" : "game-resign-other",
                                format("%s:%d", userId, slotId), user, userId);

        // /* If user drops himself, penalize him */
        // if (!dead && userId == conn.getUser() && !game.isRankingDisabled()) {
        //     handlePlayerDrop(database_connection, userId, game, slotId);
        //     handlePlayerRankChanges(database_connection, userId);
        // }

        /* Notify cron to recompute host time */
        cron.handleGameChange(gameId);
    }

    conn.sendSuccess("OK");
}

/** PLAYERLIST command. */
void
doPlayerList(HostConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCountAtLeast(1);
    int32_t gameId = toIntResult(args.getNext());
    bool allPlayers = false;
    while (args.getNumArgs() > 0) {
        string_t keyword = strUCase(toStringResult(args.getNext()));
        if (keyword == "ALL") {
            allPlayers = true;
        } else {
            conn.sendError("400 Syntax error");
            return;
        }
    }

    // Execute
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    conn.checkPermission(game, Game::ReadPermission);
    game.loadRaceNames(database_connection);

    // Adjust allPlayers
    string_t turn1State;
    if (allPlayers) {
        string_t gameState = game.getState();
        if ((gameState != "running" && gameState != "finished") || game.getConfigInt("turn") <= 0) {
            allPlayers = false;
        } else {
            turn1State = game.getGame().subtree("turn").subtree(1).hashKey("info").getStringField("turnstatus");
        }
    }

    Ptr<IntArrayData> ad = new IntArrayData();
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if ((turn1State.size() >= 2U*i && getInt16(turn1State.data() + 2*i) >= 0) || game.isSlotInGame(i)) {
            ad->content.pushNew(makeIntValue(i));
            ad->content.pushNew(game.describeSlot(i, conn.getUser()));
        }
    }
    ad->addDimension(ad->content.size());
    IntArray a(ad);
    conn.sendObject(&a);
}

/** PLAYERSTAT command. */
void
doPlayerStat(HostConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCount(2);
    int32_t gameId = toIntResult(args.getNext());
    int32_t slotId = toIntResult(args.getNext());

    // Execute
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    conn.checkPermission(game, Game::ReadPermission);

    if (!game.isSlotInGame(slotId)) {
        conn.sendError("412 Slot not in use");
        return;
    }

    game.loadRaceNames(database_connection);
    std::auto_ptr<IntValue> result(game.describeSlot(slotId, conn.getUser()));
    conn.sendObject(result.get());
}

/** PLAYERSETDIR command. */
void
doPlayerSetDir(HostConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCount(3);
    int32_t gameId    = toIntResult(args.getNext());
    string_t playerId = toStringResult(args.getNext());
    string_t dirName  = toStringResult(args.getNext());

    // Connect to servers and check permissions
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    if ((!conn.isAdmin() && conn.getUser() != playerId) || !game.isUserOnGame(playerId)) {
        conn.sendError("403 Permission denied");
        return;
    }
    ReconnectingGuard fg(userfile_rc);
    FileClient file(userfile_connection);

    // Create the new game directory. This will throw an error if we don't
    // have permissions to do that. Also make sure we don't overwrite a
    // different game's data.
    file.setUserPermissions(conn.getUser());
    if (!dirName.empty()) {
        file.createDirectoryTree(dirName);
        int32_t oldGameId = file.getIntProperty(dirName, "game");
        if (oldGameId != 0 && oldGameId != gameId) {
            conn.sendError("601 Directory in use");
            return;
        }
    }

    // Lock the game. This will throw when the game is currently hosting.
    // It will also make sure nobody is fumbling with the game in parallel.
    GameLock gl(gameId);
    gl.lockOrDie(IntentDB);

    // Get old state.
    string_t oldDir = game.getPlayerConfig(playerId, "gameDir");
    if (oldDir != dirName && !oldDir.empty()) {
        uninstallGameData(file, oldDir);
    }

    // Change game directory.
    game.setPlayerConfig(playerId, "gameDir", dirName);

    // Install game data
    if (!dirName.empty()) {
        installGameData(game, game.getSlotsByPlayer(playerId), file, dirName);
    }
    conn.sendSuccess("OK");
}

/** PLAYERGETDIR command. */
void
doPlayerGetDir(HostConnection& conn, IntArgBlock& args)
{
    args.checkArgumentCount(2);
    int32_t gameId    = toIntResult(args.getNext());
    string_t playerId = toStringResult(args.getNext());

    // Execute
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    if ((!conn.isAdmin() && conn.getUser() != playerId) || !game.isUserOnGame(playerId)) {
        conn.sendError("403 Permission denied");
        return;
    }

    IntStringValue sv(game.getPlayerConfig(playerId, "gameDir"));
    conn.sendObject(&sv);
}

/** PLAYERCHECKFILE command. */
void
doPlayerCheckFile(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCountAtLeast(3);
    int32_t gameId    = toIntResult(args.getNext());
    string_t playerId = toStringResult(args.getNext());
    string_t fileName = toStringResult(args.getNext());

    bool checkGameDir = false;
    string_t gameDirToCheck;
    while (args.getNumArgs() > 0) {
        string_t keyword = strUCase(toStringResult(args.getNext()));
        if (keyword == "DIR") {
            args.checkArgumentCountAtLeast(1);
            checkGameDir = true;
            gameDirToCheck = toStringResult(args.getNext());
        } else {
            conn.sendError("400 Syntax error");
            return;
        }
    }

    // Execute
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    if ((!conn.isAdmin() && conn.getUser() != playerId) || !game.isUserOnGame(playerId)) {
        conn.sendError("403 Permission denied");
        return;
    }

    // Check game directory. Must be present, and the same as the specified one.
    string_t gameDir = game.getPlayerConfig(playerId, "gameDir");
    if ((checkGameDir && gameDirToCheck != gameDir) || gameDir.empty()) {
        conn.sendSuccess("stale");
        return;
    }

    // Check file name.
    if (isPreciousFile(fileName)) {
        conn.sendSuccess("allow");
        return;
    }

    // Could it be a turn file?
    int slot;
    if (fileName.size() >= 11
        && fileName.compare(0, 6, "player", 6) == 0
        && fileName.compare(fileName.size()-4, 4, ".trn", 4) == 0
        && stringValue(fileName.substr(6, fileName.size()-10), slot)
        && slot > 0
        && slot <= NUM_PLAYERS)
    {
        StringList_t players;
        game.listPlayers(slot, players);
        if (std::find(players.begin(), players.end(), playerId) != players.end()) {
            conn.sendSuccess("trn");
            return;
        }
    }

    // Not permitted.
    conn.sendSuccess("refuse");
}

/** TRN command. */
void
doTurnUpload(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCountAtLeast(1);
    string_t trnData = toStringResult(args.getNext());
    string_t mail;
    string_t info;
    bool checkMail = false;
    int32_t gameLimit = 0;
    int32_t slotLimit = 0;
    while (args.getNumArgs() > 0) {
        args.checkArgumentCountAtLeast(1);
        string_t keyword = strUCase(toStringResult(args.getNext()));
        if (keyword == "GAME") {
            args.checkArgumentCountAtLeast(1);
            gameLimit = toIntResult(args.getNext());
        } else if (keyword == "SLOT") {
            args.checkArgumentCountAtLeast(1);
            slotLimit = toIntResult(args.getNext());
        } else if (keyword == "MAIL") {
            args.checkArgumentCountAtLeast(1);
            mail = toStringResult(args.getNext());
            checkMail = true;
        } else if (keyword == "INFO") {
            args.checkArgumentCountAtLeast(1);
            info = toStringResult(args.getNext());
        } else {
            conn.sendError("400 Syntax error");
            return;
        }
    }

    // Parse the turn file and complete the parameters
    MemoryStream ms(trnData);
    GTurnfile trn(ms, false);
    if (slotLimit == 0) {
        slotLimit = trn.getPlayer();
    }
    if (gameLimit == 0) {
        ReconnectingGuard g(database_rc);
        gameLimit = DbSubtree(database_connection, GAME_ROOT).subtree("bytime").intKey(trn.getTimestamp().getTimestampAsString()).get();
    }
    if (gameLimit == 0) {
        conn.sendError("404 Game does not exist");
        return;
    }

    // Lock game
    GameLock gl(gameLimit);
    gl.lockOrDie(IntentCheckturn);

    // Check permissions
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameLimit);
    StringList_t players;
    game.listPlayers(slotLimit, players);


    bool isAdmin = conn.isAdmin();
    string_t user = conn.getUser();
    uint32_t thisPlayerIndex;
    if (checkMail) {
        // Identify player by his email address. This is only allowed for admins.
        if (!isAdmin) {
            conn.sendError("403 Permission denied");
            return;
        }
        // Start with out-of-range
        isAdmin = false;
        thisPlayerIndex = players.size();
        for (uint32_t i = players.size(); i > 0; --i) {
            string_t thisPlayerMail = DbSubtree(database_connection, USER_ROOT).subtree(players[i-1]).hashKey("profile").getStringField("email");
            if (strLCase(thisPlayerMail) == strLCase(mail)) {
                // Found him, so assume his credentials.
                // (If no match is found, we'll fall through to the "Permission denied" below).
                thisPlayerIndex = i-1;
                user = players[i-1];

                MutexGuard g(output_mutex);
                std::printf("[cmd] address '%s' resolved to player '%s'\n", mail.c_str(), user.c_str());
                break;
            }
        }
    } else {
        // Use authenticated player
        thisPlayerIndex = indexOf(players, user);
    }
    if (!isAdmin && thisPlayerIndex >= players.size()) {
        if (checkMail) {
            conn.sendError("407 Mail mismatch");
        } else {
            conn.sendError("403 Permission denied");
        }
        return;
    }
    if (game.getState() != "running") {
        conn.sendError("412 Wrong game state");
        return;
    }

    // Upload turn
    string_t gameDir = exportEnvironment(database_connection, gameLimit);
    {
        ReconnectingGuard fg(file_rc);
        FileClient file(file_connection);
        file.setUserPermissions(user);
        file.putFile(format("%s/in/new/player%d.trn", gameDir, slotLimit), trnData);
    }

    // Run checkturn
    // FIXME: this keeps the database locked while running checkturn,
    // and might thus block it for a considerable amount of time.
    // Can we do better?
    IntDataSegment cmd;
    cmd.pushNew(makeStringValue("/bin/sh"));
    cmd.pushNew(makeStringValue("bin/checkturn.sh"));
    cmd.pushNew(makeStringValue(gameDir));
    cmd.pushNew(makeIntValue(slotLimit));
    string_t output;
    int32_t code = checkturn_runner.run(cmd, output);

    // Process result
    Game::Slot slot(game.getSlot(slotLimit));
    int32_t existingState = slot.turnStatus().get();
    int32_t newState = (code + 1);
    if (newState == TurnYellow
        || newState == TurnGreen
        || ((existingState == TurnMissing
             || existingState == TurnDead)
            && newState <= TurnBad))
    {
        // Turn has been kept, or there has not been a previous turn
        slot.turnStatus().set(newState);
    }

    // Log it
    std::printf("[trn] game %d, player %d, user %s, status %d, info %s\n",
                gameLimit,
                slotLimit,
                user.size() ? user.c_str() : "(none)",
                newState,
                info.size() ? info.c_str() : "(none)");

    // Update
    if (newState == TurnYellow || newState == TurnGreen) {
        // Turn has been kept, so trigger scheduler
        game.setConfigInt("lastTurnSubmitted", getCurrentTime());
        cron.handleGameChange(gameLimit);

        // Maybe we're taking over again for our replacement
        if (!isAdmin) {
            while (thisPlayerIndex+1 < players.size()) {
                game.popPlayerSlot(slotLimit, DbSubtree(database_connection, USER_ROOT));
                players.pop_back();
            }
        }

        // Distribute turn.
        ReconnectingGuard fg(userfile_rc);
        FileClient file(userfile_connection);
        installFileMulti(game, players, file, format("player%d.trn", slotLimit), trnData, slotLimit);
    }

    // Build protocol result
    Ptr<IntArrayData> a = new IntArrayData();
    a->content.pushNew(makeStringValue("status"));
    a->content.pushNew(makeIntValue(newState));
    a->content.pushNew(makeStringValue("output"));
    a->content.pushNew(makeStringValue(output));
    a->content.pushNew(makeStringValue("game"));
    a->content.pushNew(makeIntValue(gameLimit));
    a->content.pushNew(makeStringValue("slot"));
    a->content.pushNew(makeIntValue(slotLimit));
    a->content.pushNew(makeStringValue("previous"));
    a->content.pushNew(makeIntValue(existingState));
    a->content.pushNew(makeStringValue("user"));
    a->content.pushNew(makeStringValue(user));
    a->addDimension(a->content.size());
    IntArray aa(a);
    conn.sendObject(&aa);
}

/** TRNMARKTEMP command. */
void
doTurnMarkTemporary(HostConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(3);
    int32_t gameId = toIntResult(args.getNext());
    int32_t slotId = toIntResult(args.getNext());
    int32_t enable = toIntResult(args.getNext());

    // Execute
    ReconnectingGuard g(database_rc);
    Game game(DbSubtree(database_connection, GAME_ROOT), gameId);
    if (!game.isSlotInGame(slotId)) {
        conn.sendError("412 Slot not in use");
        return;
    }

    Game::Slot slot(game.getSlot(slotId));
    if (!conn.isAdmin()) {
        // Not admin: user must be playing this slot
        StringList_t players;
        slot.players().getAll(players);
        if (indexOf(players, conn.getUser()) >= players.size()) {
            conn.sendError("403 Permission denied");
            return;
        }
    }

    // Check existing turn state
    int32_t turnState = slot.turnStatus().get() & TurnStateMask;
    if (turnState == TurnGreen || turnState == TurnYellow) {
        // ok
        if (enable) {
            turnState |= TurnIsTemporary;
        }
        slot.turnStatus().set(turnState);
        cron.handleGameChange(gameId);
        conn.sendSuccess("OK");
    } else {
        // bad turn state (no turn submitted, bad turn, ...)
        conn.sendError("412 Bad turn state");
    }
}
