/**
  *  \file planetscentral/host/cron.cc
  *  \brief Host: Scheduler
  *
  *  The scheduler has three queues:
  *  - the main scheduler queue (schedule) containing future events.
  *  - a queue of overdue events (overdue). All games in this queue are locked.
  *  - a queue of games notified by other components (new_commands).
  *
  *  To support examinability, this code maintains the invariant that every
  *  game which is subject to scheduling actually appears in one of the three
  *  lists. Therefore, for example, we don't take a game out of the overdue
  *  list, run host, and place it in the schedule list; we keep it in the
  *  overdue list for the whole host run so people see that something happens
  *  there.
  */

#include <list>
#include <stdio.h>
#include "planetscentral/host/cron.h"
#include "cpluslib/misc.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/host/game.h"
#include "planetscentral/host/lock.h"
#include "planetscentral/host/schedule.h"
#include "planetscentral/host/exec.h"
#include "util/except.h"
#include "planetscentral/host/db.h"

/******************************* Utilities *******************************/

/** Sort two ScheduleItems by time. */
bool
operator<(const Cron::ScheduleItem& a, const Cron::ScheduleItem& b)
{
    if (a.time != b.time)
        return a.time < b.time;
    if (a.gameId != b.gameId)
        return a.gameId < b.gameId;
    return a.action < b.action;
}

namespace {
    /** Sort predicate to find a game in a list of ScheduleItems. */
    class IsGame {
     public:
        IsGame(int32_t gameId)
            : gameId(gameId) { }
        bool operator()(const Cron::ScheduleItem& i)
            { return i.gameId == gameId; }
     private:
        int32_t gameId;
    };
}

/** Check that all turns are in for a game. */
static bool
checkAllTurnsIn(Game& game)
{
    bool haveAnyTurns = false;
    for (int i = 1; i <= 11; ++i) {
        Game::Slot slot(game.getSlot(i));
        int32_t turnStatus = slot.turnStatus().get();
        if (turnStatus == TurnGreen || turnStatus == TurnYellow) {
            // Turn is green or yellow, that's ok for hosting.
            haveAnyTurns = true;
        } else if (turnStatus == TurnDead
                   || (turnStatus == TurnMissing
                       && (slot.slotStatus().get() == 0
                           || slot.players().size() == 0)))
        {
            // There is no turn because the slot is empty or not played.
            // That's ok for hosting, but we still need some turns for other players.
        } else {
            // Turn is missing (or temporary)
            return false;
        }
    }
    return haveAnyTurns;
}

/** Compute time for a running game. Database lock must be held. The
    schedule item will be produced on the given list, \c sch.

    \param now    [in] Current time
    \param root   [in] Game subtree root
    \param gameId [in] Game to work on
    \param sch    [out] Schedule will be produced here */
void
computeGameHostTimes(const int32_t now, DbSubtree root, const int32_t gameId, std::list<Cron::ScheduleItem>& sch)
{
    // DB objects
    Game gg(root, gameId, Game::NoExistanceCheck);
    DbSubtree g         = root.subtree(gameId);
    DbSubtree schedules = g.subtree("schedule");
    DbHash    settings  = g.hashKey("settings");
    int32_t turn = settings.getIntField("turn");

    // If this game was never hosted, we must master it before! Games will be in state "running"
    // when explicitly placed in that state, even though they have not been mastered yet.
    // So, master it immediately. We cannot check masterHasRun, because that is already set
    // for premastered games where we still must run through our master script to complete the
    // game directory.
    if (turn == 0) {
        sch.push_back(Cron::ScheduleItem(MasterAction, gameId, now));
        return;
    }

    // Start by expiring obsolete schedules
    DbStringList scheduleList = schedules.stringListKey("list");

    Schedule currentSchedule;
    bool currentScheduleValid = false;
    bool haveDroppedSchedule = false;
    while (!currentScheduleValid && scheduleList.size() > 0) {
        string_t currentScheduleId = scheduleList[0];
        currentSchedule.loadFrom(schedules.hashKey(currentScheduleId));
        if (currentSchedule.isExpired(turn, now)) {
            // This schedule is expired, drop it
            scheduleList.popFront();
            schedules.hashKey(currentScheduleId).remove();
            currentScheduleValid = false;
            haveDroppedSchedule = true;
        } else {
            // This schedule is valid
            currentScheduleValid = true;
        }
    }

    // Create a schedule expiration event
    int32_t scheduleChangeTime = 0;
    if (currentScheduleValid && currentSchedule.getCondition() == Schedule::TimeCondition) {
        scheduleChangeTime = currentSchedule.getConditionArg();
    }

    // Now compute game events.
    int32_t nextHostTime = 0;
    int32_t lastHostTime = g.hashKey("settings").getIntField("lastHostTime");

    // If we have dropped a schedule, adjust lastHostTime.
    if (haveDroppedSchedule && lastHostTime > 0 && currentScheduleValid) {
        int32_t virtualTime = currentSchedule.getPreviousVirtualHost(now);
        if (virtualTime != 0 && virtualTime > lastHostTime) {
            lastHostTime = virtualTime;
            g.hashKey("settings").setField("lastHostTime", lastHostTime);
        }
    }

    if (lastHostTime == 0) {
        // Host never ran, so schedule it for running immediately.
        nextHostTime = now;
    } else if (!currentScheduleValid || currentSchedule.getType() == Schedule::Stopped) {
        // Game is not hosted
    } else if (currentSchedule.getType() == Schedule::Quick) {
        // Game is hosted when all turns are in
        if (checkAllTurnsIn(gg)) {
            nextHostTime = g.hashKey("settings").getIntField("lastTurnSubmitted") + currentSchedule.getHostDelay();
        }
    } else if (currentSchedule.getType() == Schedule::Manual) {
        // Game is hosted when flag is set
        if (settings.intField("hostRunNow").get() != 0) {
            nextHostTime = now;
        } else if (currentSchedule.getHostEarly() && checkAllTurnsIn(gg)) {
            nextHostTime = g.hashKey("settings").getIntField("lastTurnSubmitted") + currentSchedule.getHostDelay();
        }
    } else {
        // Game is hosted on fixed schedule. Note that a schedule may also happen
        // to return time 0 (e.g. if it's a weekly schedule with no days).
        nextHostTime = currentSchedule.getNextHost(lastHostTime);
        if (nextHostTime != 0) {
            if (currentSchedule.getHostEarly() && checkAllTurnsIn(gg)) {
                int32_t accHostTime = g.hashKey("settings").getIntField("lastTurnSubmitted") + currentSchedule.getHostDelay();
                if (accHostTime < nextHostTime) {
                    nextHostTime = accHostTime;
                }
            }
        }
    }

    // Generate exactly one event.
    if (nextHostTime > 0 && (scheduleChangeTime == 0 || nextHostTime <= scheduleChangeTime)) {
        sch.push_back(Cron::ScheduleItem(HostAction, gameId, nextHostTime));
        g.hashKey("settings").setField("nextHostTime", nextHostTime);
    } else if (scheduleChangeTime > 0 && (nextHostTime == 0 || scheduleChangeTime < nextHostTime)) {
        sch.push_back(Cron::ScheduleItem(ScheduleChangeAction, gameId, scheduleChangeTime));
        g.hashKey("settings").removeField("nextHostTime");
    }
}

/** Compute master time for a joining game. Database lock must be held. The
    schedule item will be produced on the given list, \c sch.

    \param now    [in] Current time
    \param root   [in] Game subtree root
    \param gameId [in] Game to work on
    \param sch    [out] Schedule will be produced here */
void
computeGameMasterTimes(const int32_t now, DbSubtree root, const int32_t gameId, std::list<Cron::ScheduleItem>& sch)
{
    // Check whether this game should be started
    Game game(root, gameId, Game::NoExistanceCheck);
    if (!game.hasAnyOpenSlot()) {
        int32_t time = game.getConfigInt("lastPlayerJoined");
        if (time == 0) {
            time = now;
        } else {
            time += 15;  /* FIXME: make this configurable */
        }
        sch.push_back(Cron::ScheduleItem(MasterAction, gameId, time));
    }
}

/** Compute time for a game. Calls computeGameMasterTimes or computeGameHostTimes,
    depending on the game's state.

    \param now    [in] Current time
    \param root   [in] Game subtree root
    \param gameId [in] Game to work on
    \param sch    [out] Schedule will be produced here */
void
computeGameTimes(const int32_t now, DbSubtree root, const int32_t gameId, std::list<Cron::ScheduleItem>& sch)
{
    if (!root.intSetKey("broken").contains(gameId)) {
        string_t gameState = root.subtree(gameId).stringKey("state").get();
        if (gameState == "joining") {
            computeGameMasterTimes(now, root, gameId, sch);
        }
        if (gameState == "running") {
            computeGameHostTimes(now, root, gameId, sch);
        }
    }
}

/** Name an action.
    \param a The action */
const char*
getCronActionName(CronAction a)
{
    static const char*const actions[] = {
        "unknown",
        "none",
        "host",
        "schedulechange",
        "master"
    };
    if (a >= 0 && a < int(countof(actions))) {
        return actions[a];
    } else {
        return "?";
    }
}

/** Log an action.
    \param what What we'll do with this action
    \param item The action */
static void
logAction(const char* what, const Cron::ScheduleItem& item)
{
    // MutexGuard g(output_mutex);
    printf("[cron] %s: %d (%s, t=%d)\n", what, int(item.gameId), getCronActionName(item.action), item.time);
}

/********************************** Cron *********************************/

/** Constructor.
    \param hostRunner   Runner to execute the host
    \param dbConnection Database connection
    \param dbLock       Database lock */
Cron::Cron(Runner& hostRunner, DbClient& dbConnection, ReconnectingConnection& dbLock)
    : hostRunner(hostRunner),
      dbConnection(dbConnection),
      dbLock(dbLock),
      new_command_sem(0)
{ }

/** Start the cron background thread. */
void
Cron::start()
{
    pthread_t handle;
    if (pthread_create(&handle, 0, cron, this) < 0) {
        perror("unable to create 'cron' thread");
    }
}

/** Handle start of a game. Called on joining->running transition of a
    game. Runs master and the initial host. */
void
Cron::handleGameStart(int32_t id)
{
    MutexGuard g(mutex);
    new_commands.push_back(id);
    new_command_sem.up();
}

/** Handle change of a game's settings. Checks whether this means a
    new host date. */
void
Cron::handleGameChange(int32_t id)
{
    MutexGuard g(mutex);
    new_commands.push_back(id);
    new_command_sem.up();
}

/** Get next action for a game. Looks only at the schedules. */
Cron::ScheduleItem
Cron::getNextAction(int32_t gameId)
{
    MutexGuard g(mutex);

    // If game is currently under reconsideration, say that I don't know.
    if (std::find(new_commands.begin(), new_commands.end(), gameId) != new_commands.end()) {
        return ScheduleItem(UnknownAction, gameId, 0);
    }

    // Is it overdue?
    std::list<ScheduleItem>::iterator p = std::find_if(overdue.begin(), overdue.end(), IsGame(gameId));
    if (p != overdue.end()) {
        // Return time 0 to mean now
        return ScheduleItem(p->action, p->gameId, 0);
    }

    // Is it in the future?
    p = std::find_if(schedule.begin(), schedule.end(), IsGame(gameId));
    if (p != schedule.end()) {
        return *p;
    }

    // Nothing found
    return ScheduleItem(NoAction, gameId, 0);
}

/** Get all scheduler events.
    \param output [out] Scheduler events will be appended here */
void
Cron::getAllSchedules(std::list<ScheduleItem>& output)
{
    MutexGuard g(mutex);
    for (std::list<ScheduleItem>::iterator p = overdue.begin(); p != overdue.end(); ++p) {
        output.push_back(ScheduleItem(p->action, p->gameId, 0));
    }
    for (std::list<ScheduleItem>::iterator p = schedule.begin(); p != schedule.end(); ++p) {
        output.push_back(*p);
    }
}

/** Cron main loop. */
void*
Cron::cron(void* self)
{
    Cron* me = static_cast<Cron*>(self);

    // Generate initial schedule
    printf("[cron] generating initial schedule...\n");
    me->generateInitialSchedule();
    printf("[cron] generated %d events.\n", int(me->schedule.size()));

    // Main loop
    while (1) {
        // Process incoming requests
        me->processRequests();

        // Move due items to the overdue list
        me->moveDueItems();

        // Figure out what to do
        enum { Nothing, Run, Wait } what = Nothing;
        ScheduleItem item(NoAction, 0, 0);
        {
            MutexGuard g(me->mutex);
            if (!me->overdue.empty()) {
                item = me->overdue.front();
                what = Run;
            } else if (!me->schedule.empty()) {
                item = me->schedule.front();
                what = Wait;
            }
        }

        // Process it
        if (what == Wait) {
            // Wait for scheduled event
            me->new_command_sem.timedDown(getSystemTimeFromTime(item.time));
        } else if (what == Run) {
            // Execute item
            std::list<ScheduleItem> newSchedule;
            me->runDueItem(item.gameId, newSchedule);

            // Update schedules
            MutexGuard g(me->mutex);
            me->overdue.remove_if(IsGame(item.gameId));
            me->schedule.merge(newSchedule);

            // Unlock the game
            unlockGame(item.gameId);
        } else {
            // Nothing to do, wait for request
            me->new_command_sem.down();
        }
    }

    return 0;
}

/** Generate initial schedule. Examines all games, and generates their schedule.
    Since this is called once at startup, it needn't be particularily careful
    about maintaining parallelism and examinability. */
void
Cron::generateInitialSchedule()
{
    ReconnectingGuard g(dbLock);
    MutexGuard m(mutex);
    DbSubtree root(dbConnection, "game:");

    int32_t now = getCurrentTime();

    // Broken games
    IntList_t broken;
    root.intSetKey("broken").getAll(broken);
    if (!broken.empty()) {
        printf("[cron] [BROKEN] There are %d broken games.\n", broken.size());
    }

    // Running games
    {
        IntList_t list;
        root.intSetKey("state:running").getAll(list);
        for (IntList_t::iterator pi = list.begin(); pi != list.end(); ++pi) {
            if (std::find(broken.begin(), broken.end(), *pi) == broken.end()) {
                computeGameHostTimes(now, root, *pi, schedule);
            }
        }
    }

    // Joining games
    {
        IntList_t list;
        root.intSetKey("state:joining").getAll(list);
        for (IntList_t::iterator pi = list.begin(); pi != list.end(); ++pi) {
            if (std::find(broken.begin(), broken.end(), *pi) == broken.end()) {
                computeGameMasterTimes(now, root, *pi, schedule);
            }
        }
    }

    schedule.sort();
}

/** Process incoming requests. This takes game Ids from the command queue,
    recomputes their schedules, and updates the global schedule. */
void
Cron::processRequests()
{
    DbSubtree root(dbConnection, "game:");
    while (1) {
        // Get next item to consider. Ignore items that are on the overdue list.
        // (Those will be rescheduled anyway when they get off the overdue list.)
        int32_t gameId;
        while (1) {
            MutexGuard g(mutex);
            if (new_commands.empty()) {
                // No more notifications
                return;
            }
            gameId = new_commands.front();
            if (find_if(overdue.begin(), overdue.end(), IsGame(gameId)) == overdue.end()) {
                // It's not overdue, so process it
                break;
            }
            new_commands.pop_front();
        }

        // Compute new schedule
        std::list<ScheduleItem> result;
        {
            ReconnectingGuard g(dbLock);
            computeGameTimes(getCurrentTime(), root, gameId, result);
        }
        if (!result.empty()) {
            logAction("updated", result.front());
        }

        // Update global schedule: replace this game's old schedule item with a
        // new one and take this game out of the new_commands list.
        {
            MutexGuard g(mutex);
            schedule.remove_if(IsGame(gameId));
            schedule.merge(result);
            new_commands.remove(gameId);
        }
    }
}

/** Move due items from the schedule to the overdue list.
    This marks them for immediate processing. */
void
Cron::moveDueItems()
{
    MutexGuard g(mutex);
    int32_t now = getCurrentTime();
    while (!schedule.empty() && schedule.front().time <= now) {
        logAction("due", schedule.front());
        lockGame(schedule.front().gameId, IntentRunHost);
        overdue.splice(overdue.end(), schedule, schedule.begin());
    }
}

/** Run due item.
    \param gameId      [in]  Game to work on
    \param newSchedule [out] New schedule for next event will be produced here */
void
Cron::runDueItem(int32_t gameId, std::list<ScheduleItem>& newSchedule)
{
    // Check that schedule is still current (it should be because the game is locked).
    DbSubtree root(dbConnection, "game:");
    int32_t now = getCurrentTime();
    {
        ReconnectingGuard g(dbLock);
        computeGameTimes(now, root, gameId, newSchedule);

        // Remove "run host now" signalisation.
        // If this is the HostAction, we'll find that action in newSchedule and run it.
        // If this is a different action, or the game isn't actually running on a Manual
        // schedule, we'll have to remove it to avoid it leaking through to a future
        // Manual HostAction.
        root.subtree(gameId).hashKey("settings").intField("hostRunNow").remove();
    }

    if (newSchedule.empty() || newSchedule.front().time > now) {
        // Schedule is no longer current, event happens in the future.
        // This happens, for example, when
        // - this is a ScheduleChangeAction
        // - a player resigns before the game starts
        if (!newSchedule.empty()) {
            logAction("updated", newSchedule.front());
        }
        return;
    }

    // Action should be performed
    try {
        ScheduleItem& item = newSchedule.front();
        logAction("executing", newSchedule.front());
        if (item.action == HostAction) {
            runHost(hostRunner, dbConnection, dbLock, gameId);
        } else if (item.action == MasterAction) {
            runMaster(hostRunner, dbConnection, dbLock, gameId);
        }
    }
    catch (std::exception& e) {
        printf("[cron] [exception] %s\n", e.what());
        if (FileProblemException* fpe = dynamic_cast<FileProblemException*>(&e)) {
            printf("[cron] [exception] filename: %s\n", fpe->getFileName().c_str());
        }
        printf("[cron] [BROKEN] Game %d is now broken.\n", gameId);
        Game(root, gameId, Game::NoExistanceCheck).markBroken(e.what());
    }

    // Schedule next event
    newSchedule.clear();
    {
        ReconnectingGuard g(dbLock);
        computeGameTimes(now, root, gameId, newSchedule);
    }
}
