/**
  *  \file planetscentral/host/import.cc
  *  \brief Host: Data Import
  */

#include "planetscentral/host/import.h"
#include "game/rst.h"
#include "io/io.h"
#include "io/streams.h"
#include "game/struct.h"
#include "cpluslib/format.h"
#include "util/misc.h"
#include "cpluslib/string.h"
#include "io/textfile.h"

static string_t
packScore(int32_t score[NUM_PLAYERS])
{
    char packed[NUM_PLAYERS*4];
    storeInt32(packed, score, NUM_PLAYERS);
    return string_t(packed, sizeof(packed));
}

namespace {
    class ScoreReader {
     public:
        ScoreReader(DbHash& score, DbHash& description);
        void processLine(string_t line);
        void flush();

     private:
        DbHash& score;
        DbHash& description;
        string_t name;
        string_t desc;
        int32_t values[NUM_PLAYERS];
    };
}

ScoreReader::ScoreReader(DbHash& score, DbHash& description)
    : score(score), description(description), name(), desc()
{
    /* need not initialize values, because we're not going to use them anyway
       until we see a delimiter */
}

void
ScoreReader::processLine(string_t line)
{
    /* Comment or blank? */
    string_t::size_type n = line.find_first_not_of(" \t");
    if (n >= line.size() || line[n] == '#' || line[n] == ';') {
        return;
    }

    /* Section delimiter? */
    if (line[n] == '%') {
        flush();
        name = strTrim(line.substr(n+1));
        return;
    }

    /* Must be an assignment */
    string_t::size_type p = line.find('=');
    if (p == line.npos) {
        // FIXME: log it?
        return;
    }

    string_t key(strUCase(strTrim(line.substr(n, p-n))));
    string_t value(strTrim(line.substr(p+1)));
    int32_t player, score;
    if (key == "DESCRIPTION") {
        desc = value;
    } else if (key.size() > 5
               && key.compare(0, 5, "SCORE", 5) == 0
               && stringValue(key.substr(5), player)
               && player > 0 && player <= NUM_PLAYERS)
    {
        if (stringValue(value, score)) {
            values[player-1] = score;
        }
    } else {
        // FIXME?
    }
}

void
ScoreReader::flush()
{
    if (!name.empty()) {
        score.setField(name, packScore(values));
        description.setField(name, desc);
    }
    for (int i = 0; i < NUM_PLAYERS; ++i) {
        values[i] = -1;
    }
    name.clear();
    desc.clear();
}

/** Import scores for a game.
    \param gameDir Game directory
    \param score Place scores in this hash
    \param description Description hash */
void
importGameScores(string_t gameDir, DbHash score, DbHash description)
{
    int32_t capital[NUM_PLAYERS], freighters[NUM_PLAYERS], planets[NUM_PLAYERS], bases[NUM_PLAYERS], timscore[NUM_PLAYERS];

    /* Import scores from the result files. This is less work than
       parsing score.log. It is also less work than scanning
       bdata/pdata/ship.hst (which would mean that we also have to
       read auxdata.hst to handle remote control). However, we have to
       read *all* result files, because if score blanking is enabled,
       every player only sees his own score. */
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        capital[i-1] = freighters[i-1] = planets[i-1] = bases[i-1] = timscore[i-1] = -1;
        try {
            Ptr<Stream> s = Stream::openFile(format("%s/player%d.rst", gameDir, i), Stream::C_READ);
            GResultFile rst(*s);
            s->seek(rst.getSectionOffset(GResultFile::sec_Gen));
            TResultGen gen;
            getStructureT(*s, gen);

            const TGenScore& sc = gen.scores[i-1];
            capital[i-1]    = sc.num_capital;
            freighters[i-1] = sc.num_freighters;
            planets[i-1]    = sc.num_planets;
            bases[i-1]      = sc.num_bases;
            timscore[i-1]   = 120*sc.num_bases + 10*(sc.num_capital + sc.num_planets) + sc.num_freighters;
        }
        catch (...) { }
    }

    /* Store scores */
    score.setField("capital",    packScore(capital));
    score.setField("freighters", packScore(freighters));
    score.setField("planets",    packScore(planets));
    score.setField("bases",      packScore(bases));
    score.setField("timscore",   packScore(timscore));
    description.setField("capital",    string_t());
    description.setField("freighters", string_t());
    description.setField("planets",    string_t());
    description.setField("bases",      string_t());
    description.setField("timscore",   string_t());

    /* Read plug-in scores */
    Ptr<Stream> s = Stream::openFileNT(format("%s/c2score.txt", gameDir), Stream::C_READ);
    if (s) {
        TextFile tf(*s);
        ScoreReader rdr(score, description);
        string_t line;
        while (tf.getLine(line)) {
            rdr.processLine(line);
        }
        rdr.flush();
    }
}
