/**
  *  \file planetscentral/net/mimeparser.cc
  *  \brief MIME Parsing
  */

#include <cstdio>
#include <cstring>
#include "planetscentral/net/mimeparser.h"
#include "util/recode.h"
#include "planetscentral/net/encoding.h"

/** Decode encoded-words from a header.
    \param header Incoming header line
    \return Decoded header, in UTF-8. */
static string_t
decodeHeader(string_t header)
{
    string_t result;
    string_t::size_type pos = 0, n;
    while ((n = header.find("=?", pos)) != string_t::npos) {
        // Append unencoded part
        result.append(header, pos, n-pos);
        pos = n+2;

        // Found an encoded word. Format is:
        //   =?charset?encoding?text?=
        // Parse the charset.
        n = header.find('?', pos);
        if (n == string_t::npos) {
            // encoding error
            break;
        }
        string_t charset(strLCase(string_t(header, pos, n-pos)));
        pos = n+1;

        // Parse the encoding.
        n = header.find('?', pos);
        if (n == string_t::npos) {
            // encoding error
            break;
        }
        string_t encoding(strLCase(string_t(header, pos, n-pos)));
        pos = n+1;

        // Locate the terminator
        n = header.find("?=", pos);
        if (n == string_t::npos) {
            // encoding error
            break;
        }
        string_t word(header, pos, n-pos);
        pos = n+2;

        // Decode encoding
        if (encoding == "b") {
            // base-64
            word = decodeBase64(word);
        } else {
            // treat as quoted-printable
            for (string_t::size_type i = 0; i < word.size(); ++i) {
                if (word[i] == '_') {
                    word[i] = ' ';
                }
            }
            word = decodeQP(word);
        }

        // Decode charset
        if (!CharacterSet::isCharsetNameUtf8(charset)) {
            word = CharacterSet::getCharsetByName(charset).convertToUtf8(word);
        }
        result.append(word);
    }
    result.append(header.substr(pos));
    return result;
}

/** Parse part of a header, up to a semicolon.
    This interprets double-quoting, comments (parentheses) and backslash escaping.
    It stops when encountering an unquoted semicolon.
    \param value [in] String to parse
    \param pos   [in/out] Read pointer
    \return header part */
static string_t
extractPart(const string_t& value, string_t::size_type& pos)
{
    string_t result;
    enum { Raw, DQ, DQQ, Paren, ParenQ } state = Raw;
    while (pos < value.size()) {
        const char c = value[pos++];
        switch (state) {
         case Raw:
            if (c == '"') {
                state = DQ;
            } else if (c == '(') {
                state = Paren;
            } else if (c == ';') {
                goto out;
            } else {
                result += c;
            }
            break;
         case DQ:
            if (c == '\\') {
                state = DQQ;
            } else if (c == '"') {
                state = Raw;
            } else {
                result += c;
            }
            break;
         case DQQ:
            result += c;
            state = DQ;
            break;
         case Paren:
            if (c == ')') {
                state = Raw;
            } else if (c == '\\') {
                state = ParenQ;
            } else {
                // ignore
            }
            break;
         case ParenQ:
            state = Paren;
            break;
        }
    }
 out:
    return result;
}

/** Get primary value of a header. Given a value of the form
    "primary; key=value; key=value", returns "primary".
    \param v Header value
    \param Primary value */
string_t
getHeaderPrimaryValue(string_t v)
{
    string_t::size_type i = 0;
    return extractPart(v, i);
}

/** Get secondary value of a header. Given a value of the form
    "primary; key=value; key=value", returns the "value"
    corresponding to the matching key.
    \param v Header value
    \param key Key to extract
    \param Key value */
string_t
getHeaderSubValue(string_t v, string_t key)
{
    string_t::size_type i = 0;
    extractPart(v, i);
    while (i < v.size()) {
        const string_t part(strTrim(extractPart(v, i)));
        if (part.size() > key.size() && strLCase(string_t(part, 0, key.size())) == key) {
            string_t::size_type n = part.find_first_not_of(" \t", key.size());
            if (n != part.npos && part[n] == '=') {
                return strLTrim(part.substr(n+1));
            }
        }
    }
    return string_t();
}

/** Get address from a header value. */
string_t
getHeaderAddressValue(string_t v)
{
    // Parse away "()" comments
    string_t parsed = getHeaderPrimaryValue(v);

    // Now we have either "realname <add@ress>" or "add@ress" format.
    string_t::size_type n = parsed.find('<');
    string_t::size_type p;
    if (n != string_t::npos && (p = parsed.find('>', n+1)) != string_t::npos) {
        // angle-brackets format
        parsed.erase(p);
        parsed.erase(0, n+1);
    } else {
        // not angle brackets. Might still contain a "," if it is an address
        // list, so remove that.
        n = parsed.find(',');
        if (n != string_t::npos) {
            parsed.erase(n);
        }
    }

    // Final checks
    parsed = strTrim(parsed);
    if (parsed.find('@') == string_t::npos || parsed.find_first_of(" \t") != string_t::npos) {
        return string_t();
    } else {
        return parsed;
    }
}

/******************************* MimeParser ******************************/

/** Constructor. */
MimeParser::MimeParser()
    : m_Headers(),
      m_Body(),
      m_Accum(),
      m_State(StateHeader),
      m_pLastHeader(0)
{ }

/** Add data. This can be any amount of data, including multiple or partial lines.
    \param data [in] Data
    \param size [in] Number of bytes */
void
MimeParser::addData(const char* data, std::size_t size)
{
    while (size != 0) {
        std::size_t did = size;
        switch (m_State) {
         case StateHeader:
            did = addDataHeader(data, size);
            break;
         case StatePlainBody:
            did = addDataPlainBody(data, size);
            break;
         case StateBase64Body:
            did = addDataBase64Body(data, size);
            break;
         case StateQPBody:
            did = addDataQPBody(data, size);
            break;
        }
        data += did;
        size -= did;
    }
}

/** Finish. Performs any possible postprocessing. Call this after a sequence
    of addData(), before using the object. addData() may not be called after
    finish() has been called. */
void
MimeParser::finish()
{
    switch (m_State) {
     case StateHeader:
        // Two newlines to get it out of the "Header" state
        addDataHeader("\n\n", 2);
        break;
     case StatePlainBody:
        break;
     case StateBase64Body:
        flushBase64Body();
        break;
     case StateQPBody:
        // Decode pending partial line. Do not handle trailing "=",
        // because there's no line we could join this with.
        m_Body.back() += decodeQP(m_Accum);
        break;
    }
    m_Accum.clear();
}

/** Reset this object into the state it had after construction. */
void
MimeParser::clear()
{
    m_Headers.clear();
    m_Body.clear();
    m_Accum.clear();
    m_State = StateHeader;
    m_pLastHeader = 0;
}

/** Get header by name.
    \param name Name of header field, in lower-case
    \return Header value; empty string if header is missing */
string_t
MimeParser::getHeader(string_t name) const
{
    HeaderMap_t::const_iterator it = m_Headers.find(name);
    if (it != m_Headers.end()) {
        return it->second;
    } else {
        return string_t();
    }
}

/** Get file name. Looks for headers that indicate a file name of this
    MIME entity.
    \return file name; empty string if none */
string_t
MimeParser::getFileName() const
{
    // Content-Disposition=>filename
    HeaderMap_t::const_iterator it = m_Headers.find("content-disposition");
    if (it != m_Headers.end()) {
        string_t name = getHeaderSubValue(it->second, "filename");
        if (!name.empty()) {
            return name;
        }
    }

    // Content-Type=>name
    it = m_Headers.find("content-type");
    if (it != m_Headers.end()) {
        string_t name = getHeaderSubValue(it->second, "name");
        if (!name.empty()) {
            return name;
        }
    }

    return string_t();
}

/** Get entire body as string. */
string_t
MimeParser::getBodyAsString() const
{
    string_t result = m_Body[0];
    for (BodyVec_t::size_type i = 1; i < m_Body.size(); ++i) {
        result += '\n';
        result += m_Body[i];
    }
    return result;
}

/** Add data while in "Header" state.
    \param data Data
    \param size Number of bytes
    \return Number of processed bytes (1,size] */
std::size_t
MimeParser::addDataHeader(const char* data, std::size_t size)
{
    std::size_t did = 0;
    while (did < size) {
        const char c = data[did++];
        if (c == '\r') {
            // skip
        } else if (c == '\n') {
            // it's a header
            processHeader();
            break;
        } else {
            m_Accum += c;
        }
    }
    return did;
}

/** Add data while in "Plain Body" state.
    \param data Data
    \param size Number of bytes
    \return Number of processed bytes (1,size] */
std::size_t
MimeParser::addDataPlainBody(const char* data, std::size_t size)
{
    std::size_t did = 0;
    while (did < size) {
        const char c = data[did++];
        if (c == '\r') {
            // skip
        } else if (c == '\n') {
            m_Body.push_back(string_t());
        } else {
            m_Body.back() += c;
        }
    }
    return did;
}

/** Add data while in "base64 Body" state.
    \param data Data
    \param size Number of bytes
    \return Number of processed bytes (1,size] */
std::size_t
MimeParser::addDataBase64Body(const char* data, std::size_t size)
{
    // FIXME: this can be implemented without the accumulator.
    std::size_t did = 0;
    while (did < size) {
        const char c = data[did++];
        if (c == '\r') {
            // skip
        } else if (c == '\n') {
            // whole line
            flushBase64Body();
        } else {
            m_Accum += c;
        }
    }
    return did;
}

/** Add data while in "quoted-printable Body" state.
    \param data Data
    \param size Number of bytes
    \return Number of processed bytes (1,size] */
std::size_t
MimeParser::addDataQPBody(const char* data, std::size_t size)
{
    // FIXME: this can be implemented without the accumulator.
    std::size_t did = 0;
    while (did < size) {
        const char c = data[did++];
        if (c == '\r') {
            // skip
        } else if (c == '\n') {
            // whole line
            bool addNL = true;
            if (!m_Accum.empty() && m_Accum[m_Accum.size()-1] == '=') {
                addNL = false;
                m_Accum.erase(m_Accum.size()-1);
            }
            m_Body.back() += decodeQP(m_Accum);
            m_Accum.clear();
            if (addNL) {
                m_Body.push_back(string_t());
            }
        } else {
            m_Accum += c;
        }
    }
    return did;
}

void
MimeParser::flushBase64Body()
{
    const string_t decoded = decodeBase64(m_Accum);
    m_Accum.clear();

    for (string_t::size_type i = 0; i < decoded.size(); ++i) {
        if (decoded[i] == '\n') {
            m_Body.push_back(string_t());
        } else {
            m_Body.back() += decoded[i];
        }
    }
}

/** Process a line while in "Header" state. Processes the current line in m_Accum,
    and updates the internal state. This may also change to a body state. */
void
MimeParser::processHeader()
{
    if (m_Accum.empty()) {
        // Blank line, we're starting the body
        std::map<string_t,string_t>::iterator i = m_Headers.find("content-transfer-encoding");
        if (i == m_Headers.end()) {
            m_State = StatePlainBody;
        } else {
            string_t typ = strLCase(getHeaderPrimaryValue(i->second));
            if (typ == "base64") {
                m_State = StateBase64Body;
            } else if (typ == "quoted-printable") {
                m_State = StateQPBody;
            } else {
                if (typ != "8bit" && typ != "7bit") {
                    std::printf("[parser] unknown Content-Transfer-Encoding '%s'\n", typ.c_str());
                }
                m_State = StatePlainBody;
            }
        }
        m_Body.push_back(string_t());
    } else if (m_Accum[0] == ' ' || m_Accum[0] == '\t') {
        // Continuation.
        // RFC822: single whitespace can be replaced by CRLF + many whitespace
        // RFC2822: single whitespace can be replaced by CRLF + single whitespace
        // In any case, the whitespace stays part of the field.
        if (m_pLastHeader != 0) {
            *m_pLastHeader += decodeHeader(m_Accum);
        }
    } else {
        // Might be header line
        // Ignore non-header lines, those most likely are the Unix mbox "From -" line.
        string_t::size_type i = m_Accum.find_first_of(": \t");
        if (i != m_Accum.npos && m_Accum[i] == ':') {
            m_pLastHeader = &m_Headers[strLCase(string_t(m_Accum, 0, i))];
            *m_pLastHeader = decodeHeader(strLTrim(m_Accum.substr(i+1)));
        }
    }
    m_Accum.clear();
}


/************************** MimeMultipartParser **************************/

/** Constructor. Makes a blank object. */
MimeMultipartParser::MimeMultipartParser()
    : m_pParser(0),
      m_Boundary(),
      m_Position(0)
{ }

/** Initialize.
    \param p [in] Original message; must live as long as the MimeMultipartParser.
             Not const to avoid calling it with a temporary.
    \retval true This was a multipart message; use readPart() to read it
    \retval false This was not a multipart message */
bool
MimeMultipartParser::init(MimeParser& p)
{
    // Must have appropriate Content-Type
    MimeParser::HeaderMap_t::const_iterator i = p.getHeaders().find("content-type");
    if (i == p.getHeaders().end()) {
        return false;
    }

    string_t ctype = strLCase(getHeaderPrimaryValue(i->second));
    if (ctype.size() < 10 || ctype.compare(0, 10, "multipart/", 10) != 0) {
        return false;
    }

    // Boundary must not be empty
    string_t boundary = getHeaderSubValue(i->second, "boundary");
    if (boundary.empty()) {
        return false;
    }

    // OK, remember it
    m_pParser  = &p;
    m_Boundary = "--" + boundary;
    m_Position = 0;
    findNextBoundary();
    return true;
}

/** Read a part.
    \param p [out] Part will be stored here
    \retval true Found another part
    \retval false No more parts; %part has not been changed */
bool
MimeMultipartParser::readPart(MimeParser& part)
{
    // Find extent of part
    MimeParser::BodyVec_t::size_type pos = m_Position;
    if (!findNextBoundary()) {
        return false;
    }

    // Decode it
    part.clear();
    bool first = true;
    const MimeParser::BodyVec_t& b = m_pParser->getBody();
    while (pos < m_Position-1) {
        if (!first) {
            part.addData("\n", 1);
        }
        part.addData(b[pos].data(), b[pos].size());
        ++pos;
        first = false;
    }
    part.finish();
    return true;
}

/** Find next boundary. Advances %m_Position to point after the next boundary
    (first line of next part).
    \retval true success
    \retval false No more boundaries found */
bool
MimeMultipartParser::findNextBoundary()
{
    MimeParser::BodyVec_t::size_type pos = m_Position;
    const MimeParser::BodyVec_t& b = m_pParser->getBody();
    while (pos < b.size()) {
        if (b[pos].size() >= m_Boundary.size()
            && b[pos].compare(0, m_Boundary.size(), m_Boundary) == 0
            && b[pos].find_first_not_of('-', m_Boundary.size()) == string_t::npos)
        {
            // Line contains boundary, followed by nothing else than "-"'s.
            m_Position = pos+1;
            return true;
        }
        ++pos;
    }
    return false;
}
