/**
  *  \file planetscentral/net/smtp.cc
  *  \brief Networking: SMTP
  *  \author Stefan Reuther
  *
  *  References:
  *
  *  RFC  821  SMTP
  *  RFC  822  Mail message format
  *  RFC 1321  MD5
  *  RFC 1869  ESMTP
  *              - specifies the EHLO command, and responses of the form
  *                         250-servername ESMTP hi there
  *                         250-FOO
  *                         250-BAR
  *                         250 HUHU
  *  RFC 2104  HMAC-MD5
  *  RFC 2195  CRAM-MD5
  *  RFC 2554  SMTP-Auth
  *  RFC 2595  AUTH PLAIN


 {smtp to mail.gmx.net} Verbindung aufbauen...
 TClientSocketSMTP.OnLookup from Hamster to mail.gmx.net:smtp use mail.gmx.net:25
 TClientSocketSMTP.OnConnecting from Hamster to mail.gmx.net:smtp use mail.gmx.net:25
 TClientSocketSMTP.OnConnect from Hamster to mail.gmx.net:smtp use mail.gmx.net:25
 TClientSocketSMTP > EHLO royale
 TClientSocketSMTP < 250 STARTTLS
 TClientSocketSMTP > AUTH CRAM-MD5
 TClientSocketSMTP < 334 PDYwNDguMTE2MzMzMjIwNkBtcDAxOT4=
 TClientSocketSMTP > [...authorization...]
 TClientSocketSMTP < 235 2.7.0 Go ahead {mp019}
 {smtp to mail.gmx.net} Verschicke E-Mail "53363.msg"
 [mail.gmx.net, SMTP] SendMail 53363.msg
 TClientSocketSMTP > MAIL FROM:<streu@gmx.de>
 TClientSocketSMTP < 250 2.1.0 ok {mp019}
 TClientSocketSMTP > RCPT TO:<zap>
 TClientSocketSMTP < 250 2.1.5 ok {mp019}
 TClientSocketSMTP > DATA
 TClientSocketSMTP < 354 mail.gmx.net Go ahead {mp019}
 TClientSocketSMTP > .
 TClientSocketSMTP < 250 2.6.0 Message accepted {mp019}
 E-Mail 53363.msg wurde verschickt.
 {smtp to mail.gmx.net} Verbindung abbauen...
 TClientSocketSMTP > QUIT
 TClientSocketSMTP < 221 2.0.0 GMX Mailservices {mp019}
  */

#include <cstdio>
#include <stdexcept>
#include "planetscentral/net/smtp.h"
#include "planetscentral/net/connstream.h"
#include "planetscentral/net/mimebuilder.h"

static string_t smtp_server;
static string_t smtp_port;
static string_t smtp_from_addr;
static string_t smtp_fqdn;

bool
Smtp::checkConfig(string_t key, string_t value)
{
    if (key == "SMTP.HOST") {
        smtp_server = value;
        return true;
    } else if (key == "SMTP.PORT") {
        smtp_port = value;
        return true;
    } else if (key == "SMTP.FROM") {
        smtp_from_addr = value;
        return true;
    } else if (key == "SMTP.FQDN") {
        smtp_fqdn = value;
        return true;
    } else {
        return false;
    }
}

string_t
Smtp::getFrom()
{
    return smtp_from_addr;
}

string_t
Smtp::getFqdn()
{
    // figure out our host name
    if (smtp_fqdn.empty()) {
        char hostname[256];
        if (gethostname(hostname, sizeof(hostname)) == 0) {
            hostname[sizeof(hostname)-1] = 0;
            smtp_fqdn = hostname;
        } else {
            smtp_fqdn = "localhost";
        }
    }
    return smtp_fqdn;
}

/** Read reply from server.
    \param expect result class to expect (i.e. 3 to expect a "3xx" reply) */
void
Smtp::readReply(int expect)
{
    flush();
    string_t line;
    while (!isAtEnd()) {
        readLine(line);
        std::printf("[smtp] < %s\n", line.c_str());
        if (line.size() >= 4 && line[3] != '-') {
            int code = std::atoi(line.c_str()) / 100;
            if (code != expect) {
                throw std::runtime_error("Unexpected response from server");
            }
            return;
        }
        line.clear();
    }
    throw std::runtime_error("Server closed connection unexpectedly");
}

void
Smtp::write(const string_t& command)
{
    if (command.size() > 2) {
        // command always ends with \r\n, don't log that.
        std::printf("[smtp] > %.*s\n", command.size()-2, command.data());
    }
    send(command.data(), command.size());
}

/** Send mail via server.
    \param from     [in] Our address for SMTP envelope
    \param to       [in] Receiver address for SMTP envelope
    \param content  [in] Mail data */
void
Smtp::sendMail(string_t from, string_t to, MimeBuilder& content)
{
    // connect; throws on error
    connectRetry(smtp_server, smtp_port);

    // receive greeting
    readReply(2);

    // hello
    string_t command = "HELO ";
    command += getFqdn();
    command += "\015\012";
    write(command);
    readReply(2);

    // mail from
    command = "MAIL FROM:<" + from + ">\015\012";
    write(command);
    readReply(2);

    // rcpt to
    command = "RCPT TO:<" + to + ">\015\012";
    write(command);
    readReply(2);

    // data
    command = "DATA\015\012";
    write(command);
    readReply(3);

    // send the mail
    {
        ConnectionStream ss(*this);
        content.write(ss, true);
    }
    command = ".\015\012";
    write(command);
    readReply(2);

    // close
    command = "QUIT\015\012";
    write(command);
    readReply(2);

    close();
}
